from dataclasses import field as dataclasses_field
from dataclasses import make_dataclass

import pydantic
import strawberry
from strawberry.field import StrawberryField

from .types import FieldError


def create_root_type(fields):
    cls = make_dataclass(
        "RootType",
        fields=[
            (
                field._field_definition.origin_name,
                field._field_definition.type,
                dataclasses_field(default=field),
            )
            for field in fields
        ],
    )

    return strawberry.type(cls)


def create_validation_error_type(prefix: str, type_: StrawberryField):
    @classmethod
    def from_validation_error(cls, validation_error: pydantic.ValidationError):
        errors = validation_error.errors()
        payload = {}

        for error in errors:
            field = error["loc"][0]
            message = error["msg"]
            error_type = error["type"]

            errors = payload.setdefault(field, [])
            errors.append(FieldError(message=message, type=error_type))

        return cls(errors=type_(**payload))

    cls = make_dataclass(f"{prefix}ValidationError", [("errors", type_)])
    cls.from_validation_error = from_validation_error
    return strawberry.type(cls)



from asgi_lifespan import LifespanManager
from httpx import AsyncClient
from main import app
from ward import fixture


@fixture
async def testclient():
    async with LifespanManager(app):
        async with AsyncClient(app=app, base_url="http://testserver") as client:
            yield client



from pythonit_toolkit.pastaporto.entities import Credential
from strawberry.permission import BasePermission


class IsAuthenticated(BasePermission):
    message = "Not authenticated"

    def has_permission(self, source, info, **kwargs):
        return Credential.AUTHENTICATED in info.context.request.auth.scopes



from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Optional

import jwt

from .exceptions import InvalidPastaportoError
from .tokens import decode_pastaporto


class Credential(str, Enum):
    AUTHENTICATED = "authenticated"
    STAFF = "staff"

    def __str__(self) -> str:
        return str.__str__(self)


class RequestAuth:
    scopes: list[Credential]
    pastaporto: Pastaporto

    def __init__(self, pastaporto: Pastaporto):
        self.scopes = pastaporto.credentials
        self.pastaporto = pastaporto


@dataclass
class PastaportoUserInfo:
    id: int
    email: str

    @classmethod
    def from_data(cls, data: dict[str, Any]):
        return cls(id=data["id"], email=data["email"])


@dataclass
class Pastaporto:
    user_info: Optional[PastaportoUserInfo] = None
    credentials: list[Credential] = field(default_factory=list)

    @classmethod
    def from_token(cls, token: str, secret: str):
        try:
            decoded_token = decode_pastaporto(token, secret)
        except (
            ValueError,
            UnicodeDecodeError,
            jwt.ExpiredSignatureError,
            jwt.DecodeError,
            jwt.InvalidAudienceError,
        ):
            raise InvalidPastaportoError()

        user_info = decoded_token.get("userInfo")
        return cls(
            user_info=PastaportoUserInfo.from_data(user_info) if user_info else None,
            credentials=[
                Credential(credential) for credential in decoded_token["credentials"]
            ],
        )
